/*****************************************************************************
 *
 *   CtxH264.h
 *
 *   H.264 v3 interface software fallback implementation.
 *
 *   $Id: //icaclient/unix13.2/client/linux/obj.h264/ctxh264.h#1 $
 *
 *   Copyright 2013-2014 Citrix Systems, Inc.  All Rights Reserved.
 *
 *****************************************************************************/


#ifndef _CTX_H264_H_
#define _CTX_H264_H_

#include <vector>
#include <map>
#include <deque>

#include <pthread.h>

#ifdef X11_SUPPORT
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#endif

#include "H264_decode.h"

#define MAX_DIRTY_RECT  31 /* Taken from tw2proto.h. */

#ifndef max
#define max(a,b) (((a) > (b)) ? (a) : (b)) 
#endif
#ifndef min
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

typedef struct image_buf ImgBuf;

/* Taken from Xv.h and Xvlib.h.
 * TODO: Update build system with libxv-dev package and remove this thrid party
 *       source.
 *
 * The following XV_H and XVLIB_H #ifndef secions have the following license.
 */

/***********************************************************
Copyright 1991 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

#ifdef X11_SUPPORT
#ifndef XV_H
typedef XID XvPortID;

#define XvVideoMask	 0x00000004
#define XvStillMask	 0x00000008
#define XvImageMask	 0x00000010
#endif /* XV_H */

#ifndef XVLIB_H
typedef struct {
    char depth;
    unsigned long visual_id;
} XvFormat;

typedef struct {
    XvPortID base_id;
    unsigned long num_ports;
    char type;
    char *name;
    unsigned long num_formats;
    XvFormat *formats;
    unsigned long num_adaptors;
} XvAdaptorInfo;

typedef struct {
    int id; 
    int width;
    int height; 
    int data_size; 
    int num_planes; 
    int *pitches; 
    int *offsets; 
    char *data; 
    XPointer obdata;
} XvImage;

typedef struct {
    int id;                      /* Unique descriptor for the format */
    int type;                    /* XvRGB, XvYUV */
    int byte_order;              /* LSBFirst, MSBFirst */
    char guid[16];               /* Globally Unique IDentifier */
    int bits_per_pixel;
    int format;                  /* XvPacked, XvPlanar */
    int num_planes;

    /* for RGB formats */
    int depth;
    unsigned int red_mask;       
    unsigned int green_mask;   
    unsigned int blue_mask;   

    /* for YUV formats */
    unsigned int y_sample_bits;
    unsigned int u_sample_bits;
    unsigned int v_sample_bits;   
    unsigned int horz_y_period;
    unsigned int horz_u_period;
    unsigned int horz_v_period;
    unsigned int vert_y_period;
    unsigned int vert_u_period;
    unsigned int vert_v_period;
    char component_order[32];    /* eg. UYVY */
    int scanline_order;          /* XvTopToBottom, XvBottomToTop */
} XvImageFormatValues; 
#endif /* XVLIB_H */
#endif /* X11_SUPPORT */

typedef struct _V3_FBCtxData
{
    unsigned int    CtxId;

    /* Decoder defined by source. */

    struct Decoder  dec;

    unsigned int    uEncFrameCurSize;
    unsigned int    uEncFrameSize;
    unsigned int    uEncMaxFrameSize;
    unsigned char   *pEncodedFrame;
    unsigned int    uCtxWidth;
    unsigned int    uCtxHeight;

    pthread_mutex_t H264DataLock;

#ifdef X11_SUPPORT
    Window          window;
    XImage         *fb;
    XShmSegmentInfo shm_info;
    int             shm_id;

    /* XVideo extension stuff. */

    XvAdaptorInfo  *ai;
    XvImage        *fb_yuv;
    unsigned int    n_adaptors;
    int             xv_port;
#endif

    /* Dirty rects. */

    SIGNED_RECT     dirty_rects[MAX_DIRTY_RECT];
    int             num_rects;

    /* Some state variables to help improve decision making. */

    bool            small_frame;    /* Small frame? */
    bool            text_frame;     /* Lossless text in frame? */        
    bool            h264_frame;     /* H.264 in frame? */

    std::vector<struct image_buf> ltdata; /* Lossless text data. */
    std::vector<struct image_buf> sfdata; /* Small frame data. */

#ifdef DUMP_OUTPUT
    /* For H.264 dump output. */
    FILE            *fp;
#endif

} V3FBCtxData;

typedef std::map<H264_context, V3FBCtxData*> GstContextMap;
typedef GstContextMap::iterator GstContextMapIter;

extern struct H264_decoder H264_decoder; /* Starting point for plugin. */

bool v3_init();
H264_context v3_open_context(int width, int height, void *codec_data,
                             int len, unsigned int options);

bool v3_start_frame(H264_context Ctx, unsigned int encoded_size,
                    SIGNED_RECT dirty_rects[], unsigned int num_rects);
bool v3_decode_frame(H264_context Ctx, void* H264_data, int len, bool last);

bool v3_compose_with_fb(H264_context Ctx, ImgBuf *fb,
                        SIGNED_RECT interesting_rects[],
                        unsigned int num_rects);
bool v3_compose_with_rects(H264_context Ctx, ImgBuf text_rects[],
                           unsigned int num_rects, bool last);

bool v3_push_frame(H264_context cxt, struct window_info windows[],
                   unsigned int num_windows, bool wait, bool *pushed);

void v3_close_context(H264_context Ctx);
void v3_end ();

/* All exported by the main process. */

extern "C" char        *getICARoot(void);
extern "C" struct mainloop_plugin Plug; /* FIXME */

#ifdef X11_SUPPORT
extern "C" Display     *GetICADisplay();
extern "C" void         WaitForShmCompletion(); 
extern "C" UINT32       SHM_FramesPushed;
extern "C" int          Get_SHM_event_base();

#ifdef CLIENT_INSTRUMENTATION
extern "C" UINT32       cl_inst_flags;
#endif

/* Xv* function prototypes. */

typedef XvImage
    *(*pfn_XvShmCreateImage)(Display *dpy, XvPortID port, int id, char *data,
			     int width, int height, XShmSegmentInfo *shminfo);
typedef int
    (*pfn_XvShmPutImage)(Display *dpy, XvPortID port, Drawable d, GC gc,
			 XvImage *image, int src_x, int src_y,
			 unsigned int src_w, unsigned int src_h,
			 int dest_x, int dest_y,
			 unsigned int dest_w, unsigned int dest_h,
			 Bool send_event);
typedef int
    (*pfn_XvQueryAdaptors)(Display *dpy, Window window,
			   unsigned int *p_num_adaptors,
			   XvAdaptorInfo **pp_adaptor_info);
typedef void (*pfn_XvFreeAdaptorInfo)(XvAdaptorInfo *p_adaptor_info);
typedef XvImageFormatValues
    *(*pfn_XvListImageFormats)(Display *dpy, XvPortID port, int *count);
typedef int (*pfn_XvGrabPort)(Display *dpy, XvPortID port, Time time);

/* Common functions, for X11 support. */

bool X11_init(void);

#endif /* X11_SUPPORT */

/* Common functions, generic. */

V3FBCtxData *GetCtxData(H264_context Ctx);
void         close_context_common(V3FBCtxData *pCtxData);
V3FBCtxData *open_context_common(int width, int height, void *codec_data,
				 int len, unsigned int options);
void         v3_end(void);

/* Decoder-specific entry points, called by common functions. */

void close_context(V3FBCtxData *pCtxData);
#endif // _CTX_H264_H_
